@@ -31,6 +31,7 @@
 #include <errno.h>
 #include <stdlib.h>
 #include <signal.h>
+#include <time.h>
 #include <string.h>
 #include <unistd.h>
 #include <sys/types.h>
@@ -48,52 +49,141 @@
 #include "dbus.h"
 
 /* Default MTU's */
-#define RX_MTU 32767
-#define TX_MTU 32767
+#define DEFAULT_RX_MTU 32767
+#define DEFAULT_TX_MTU 32767
 
-#define TARGET_SIZE	16
-static const guint8 FTP_TARGET[TARGET_SIZE] = { 0xF9, 0xEC, 0x7B, 0xC4,
-					0x95, 0x3C, 0x11, 0xD2,
-					0x98, 0x4E, 0x52, 0x54,
-					0x00, 0xDC, 0x9E, 0x09 };
+#define TARGET_SIZE 16
+
+static const guint8 FTP_TARGET[TARGET_SIZE] = {
+			0xF9, 0xEC, 0x7B, 0xC4,  0x95, 0x3C, 0x11, 0xD2,
+			0x98, 0x4E, 0x52, 0x54,  0x00, 0xDC, 0x9E, 0x09  };
+
+static const guint8 PBAP_TARGET[TARGET_SIZE] = {
+			0x79, 0x61, 0x35, 0xF0,  0xF0, 0xC5, 0x11, 0xD8,
+			0x09, 0x66, 0x08, 0x00,  0x20, 0x0C, 0x9A, 0x66  };
 
 /* Connection ID */
 static guint32 cid = 0x0000;
 
 typedef struct {
-    guint8	version;
-    guint8	flags;
-    guint16	mtu;
+	guint8  version;
+	guint8  flags;
+	guint16 mtu;
 } __attribute__ ((packed)) obex_connect_hdr_t;
 
 struct obex_commands opp = {
 	.get		= opp_get,
-	.chkput		= opp_chkput,
 	.put		= opp_put,
-	.setpath	= NULL,
+	.chkput		= opp_chkput,
 };
 
 struct obex_commands ftp = {
 	.get		= ftp_get,
 	.put		= ftp_put,
+	.chkput		= ftp_chkput,
 	.setpath	= ftp_setpath,
 };
 
+struct obex_commands pbap = {
+	.get		= pbap_get,
+};
+
+static void os_reset_session(struct obex_session *os)
+{
+	if (os->fd > 0) {
+		close(os->fd);
+		os->fd = -1;
+		if (os->aborted && os->cmd == OBEX_CMD_PUT && os->current_folder) {
+			gchar *path;
+			path = g_build_filename(os->current_folder, os->name, NULL);
+			unlink(path);
+			g_free(path);
+		}
+	}
+	if (os->name) {
+		g_free(os->name);
+		os->name = NULL;
+	}
+	if (os->type) {
+		g_free(os->type);
+		os->type = NULL;
+	}
+	if (os->buf) {
+		g_free(os->buf);
+		os->buf = NULL;
+	}
+	os->aborted = FALSE;
+	os->offset = 0;
+	os->size = OBJECT_SIZE_DELETE;
+}
+
+static void os_session_mark_aborted(struct obex_session *os)
+{
+	/* the session was alredy cancelled/aborted */
+	if (os->aborted)
+		return;
+
+	os->aborted = os->size == OBJECT_SIZE_UNKNOWN ? FALSE :
+							os->size != os->offset;
+}
+
 static void obex_session_free(struct obex_session *os)
 {
-	if (os->name)
-		g_free(os->name);
-	if (os->type)
-		g_free(os->type);
+	os_reset_session(os);
+
 	if (os->current_folder)
 		g_free(os->current_folder);
-	if (os->buf)
-		g_free(os->buf);
-	if (os->fd > 0)
-		close(os->fd);
+
 	g_free(os);
 }
 
+/* From Imendio's GnomeVFS OBEX module (om-utils.c) */
+static time_t parse_iso8610(const gchar *val, int size)
+{
+	time_t time, tz_offset = 0;
+	struct tm tm;
+	gchar *date;
+	gchar tz;
+	int nr;
+
+	memset(&tm, 0, sizeof(tm));
+	/* According to spec the time doesn't have to be null terminated */
+	date = g_strndup(val, size);
+	nr = sscanf(date, "%04u%02u%02uT%02u%02u%02u%c",
+			&tm.tm_year, &tm.tm_mon, &tm.tm_mday,
+			&tm.tm_hour, &tm.tm_min, &tm.tm_sec,
+			&tz);
+	g_free(date);
+	if (nr < 6) {
+		/* Invalid time format */
+		return -1;
+	}
+
+	tm.tm_year -= 1900;	/* Year since 1900 */
+	tm.tm_mon--;		/* Months since January, values 0-11 */
+	tm.tm_isdst = -1;	/* Daylight savings information not avail */
+
+#if defined(HAVE_TM_GMTOFF)
+	tz_offset = tm.tm_gmtoff;
+#elif defined(HAVE_TIMEZONE)
+	tz_offset = -timezone;
+	if (tm.tm_isdst > 0)
+		tz_offset += 3600;
+#endif
+
+	time = mktime(&tm);
+	if (nr == 7) {
+		/*
+		 * Date/Time was in localtime (to remote device)
+		 * already. Since we don't know anything about the
+		 * timezone on that one we won't try to apply UTC offset
+		 */
+		time += tz_offset;
+	}
+
+	return time;
+}
+
 static void cmd_connect(struct obex_session *os,
 			obex_t *obex, obex_object_t *obj)
 {
@@ -124,24 +214,48 @@ static void cmd_connect(struct obex_session *os,
 	/* connection id will be used to track the sessions, even for OPP */
 	os->cid = ++cid;
 
-	register_transfer(os->cid);
+	while (OBEX_ObjectGetNextHeader(obex, obj, &hi, &hd, &hlen)) {
+		if (hi != OBEX_HDR_TARGET || hlen != TARGET_SIZE)
+			continue;
 
-	if (os->target == NULL) {
-		/* OPP doesn't contains target or connection id. */
-		OBEX_ObjectSetRsp(obj, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
-		return;
-	}
+		if (memcmp(hd.bs, FTP_TARGET, TARGET_SIZE) == 0 &&
+				os->server->services &
+						(OBEX_FTP | OBEX_PCSUITE)) {
+			os->target = FTP_TARGET;
+			os->cmds = &ftp;
+			break;
+		}
 
-	hi = hlen = 0;
-	OBEX_ObjectGetNextHeader(obex, obj, &hi, &hd, &hlen);
+		if (memcmp(hd.bs, PBAP_TARGET, TARGET_SIZE) == 0 &&
+				os->server->services & OBEX_PBAP) {
+			os->target = PBAP_TARGET;
+			os->cmds = &pbap;
+			break;
+		}
 
-	if (hi != OBEX_HDR_TARGET || hlen != TARGET_SIZE
-			|| memcmp(os->target, hd.bs, TARGET_SIZE) != 0) {
+		error("Connect attempt to a non-supported target");
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 		return;
 	}
 
+	if (os->target == NULL) {
+		if (os->server->services & OBEX_OPP) {
+			register_transfer(os->cid, os);
+			/* OPP doesn't contains target or connection id. */
+			OBEX_ObjectSetRsp(obj, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
+		} else {
+			error("Object Push connect attempt to a non-OPP server");
+			OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
+		}
+
+		return;
+	}
+
+	register_session(cid, os);
+	emit_session_created(cid);
+
 	/* Append received UUID in WHO header */
+	hd.bs = os->target;
 	OBEX_ObjectAddHeader(obex, obj,
 			OBEX_HDR_WHO, hd, TARGET_SIZE,
 			OBEX_FL_FIT_ONE_PACKET);
@@ -151,7 +265,6 @@ static void cmd_connect(struct obex_session *os,
 			OBEX_FL_FIT_ONE_PACKET);
 
 	OBEX_ObjectSetRsp(obj, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
-
 }
 
 static gboolean chk_cid(obex_t *obex, obex_object_t *obj, guint32 cid)
@@ -164,7 +277,7 @@ static gboolean chk_cid(obex_t *obex, obex_object_t *obj, guint32 cid)
 
 	os = OBEX_GetUserData(obex);
 
-	/* OPUSH doesn't provide a connection id. */
+	/* Object Push doesn't provide a connection id. */
 	if (os->target == NULL)
 		return TRUE;
 
@@ -190,7 +303,10 @@ static void cmd_get(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 	guint hlen;
 	guint8 hi;
 
-	if (!os->cmds->get) {
+	if (!os->cmds) {
+		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
+		return;
+	} else if (!os->cmds->get) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
 				OBEX_RSP_NOT_IMPLEMENTED);
 		return;
@@ -264,7 +380,10 @@ static void cmd_setpath(struct obex_session *os,
 	guint32 hlen;
 	guint8 hi;
 
-	if (!os->cmds->setpath) {
+	if (!os->cmds) {
+		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
+		return;
+	} else if (!os->cmds->setpath) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
 				OBEX_RSP_NOT_IMPLEMENTED);
 		return;
@@ -278,33 +397,26 @@ static void cmd_setpath(struct obex_session *os,
 	}
 
 	while (OBEX_ObjectGetNextHeader(obex, obj, &hi, &hd, &hlen)) {
-		if (hi == OBEX_HDR_NAME) {
-			if (os->name) {
-				debug("Ignoring multiple name headers");
-				break;
-			}
+		if (hi != OBEX_HDR_NAME)
+			continue;
 
-			/*
-			 * This is because OBEX_UnicodeToChar() accesses
-			 * the string even if its size is zero
-			 */
-			if (hlen == 0) {
-				os->name = g_strdup("");
-				break;
-			}
-
-			os->name = g_convert((const gchar *) hd.bs, hlen,
-					"UTF8", "UTF16BE", NULL, NULL, NULL);
-
-			debug("Set path name: %s", os->name);
+		if (os->name) {
+			debug("Ignoring multiple name headers");
 			break;
 		}
-	}
 
-	if (!os->name) {
-		OBEX_ObjectSetRsp(obj, OBEX_RSP_BAD_REQUEST,
-				OBEX_RSP_BAD_REQUEST);
-		return;
+		/* This is because OBEX_UnicodeToChar() accesses the string
+		   even if its size is zero */
+		if (hlen == 0) {
+			os->name = g_strdup("");
+			break;
+		}
+
+		os->name = g_convert((const gchar *) hd.bs, hlen,
+				"UTF8", "UTF16BE", NULL, NULL, NULL);
+
+		debug("Set path name: %s", os->name);
+		break;
 	}
 
 	os->cmds->setpath(obex, obj);
@@ -326,7 +438,7 @@ gboolean os_prepare_get(struct obex_session *os, gchar *file, guint32 *size)
 	os->offset = 0;
 
 	if (stats.st_size > 0)
-		os->buf = g_new0(guint8, os->tx_mtu);
+		os->buf = g_malloc0(os->tx_mtu);
 
 	*size = stats.st_size;
 
@@ -339,27 +451,43 @@ fail:
 	return FALSE;
 }
 
-static gint obex_write(struct obex_session *os,
+static gint obex_write_stream(struct obex_session *os,
 			obex_t *obex, obex_object_t *obj)
 {
 	obex_headerdata_t hd;
 	gint32 len;
+	guint8 *ptr;
 
-	debug("obex_write name: %s type: %s tx_mtu: %d fd: %d",
-			os->name, os->type, os->tx_mtu, os->fd);
+	debug("obex_write_stream: name=%s type=%s tx_mtu=%d fd=%d",
+		os->name ? os->name : "", os->type ? os->type : "",
+		os->tx_mtu, os->fd);
 
-	if (os->fd < 0)
-		return -EIO;
+	if (os->aborted)
+		return -EPERM;
+
+	if (os->fd < 0) {
+		if (os->buf == NULL)
+			return -EIO;
+
+		len = MIN(os->size - os->offset, os->tx_mtu);
+		ptr = os->buf + os->offset;
+		goto add_header;
+	}
 
 	len = read(os->fd, os->buf, os->tx_mtu);
 	if (len < 0) {
 		gint err = errno;
 		error("read(): %s (%d)", strerror(err), err);
 		g_free(os->buf);
+		os->buf = NULL;
 		return -err;
 	}
 
-	os->offset += len;
+	ptr = os->buf;
+
+add_header:
+
+	hd.bs = ptr;
 
 	if (len == 0) {
 		OBEX_ObjectAddHeader(obex, obj, OBEX_HDR_BODY, hd, 0,
@@ -369,20 +497,71 @@ static gint obex_write(struct obex_session *os,
 		return len;
 	}
 
-	hd.bs = os->buf;
+	os->offset += len;
+
 	OBEX_ObjectAddHeader(obex, obj, OBEX_HDR_BODY, hd, len,
 				OBEX_FL_STREAM_DATA);
 
 	return len;
 }
 
-static gint obex_read(struct obex_session *os,
-			obex_t *obex, obex_object_t *obj)
+gint os_prepare_put(struct obex_session *os)
+{
+	gchar *path;
+	gint len;
+
+	path = g_build_filename(os->current_folder, os->name, NULL);
+
+	os->fd = open(path, O_WRONLY | O_CREAT, 0600);
+	if (os->fd < 0) {
+		error("open(%s): %s (%d)", path, strerror(errno), errno);
+		g_free(path);
+		return -EPERM;
+	}
+
+	g_free(path);
+
+	if (os->target == NULL)
+		emit_transfer_started(os->cid);
+
+	if (!os->buf) {
+		debug("PUT request checked, no buffered data");
+		return 0;
+	}
+
+	len = 0;
+	while (len < os->offset) {
+		gint w;
+
+		w = write(os->fd, os->buf + len, os->offset - len);
+		if (w < 0) {
+			gint err = errno;
+			if (err == EINTR)
+				continue;
+			else
+				return -err;
+		}
+
+		len += w;
+	}
+
+	return 0;
+}
+
+static gint obex_read_stream(struct obex_session *os, obex_t *obex,
+				obex_object_t *obj)
 {
 	gint size;
 	gint32 len = 0;
 	const guint8 *buffer;
 
+	if (os->aborted)
+		return -EPERM;
+
+	/* workaround: client didn't send the object lenght */
+	if (os->size == OBJECT_SIZE_DELETE)
+		os->size = OBJECT_SIZE_UNKNOWN;
+
 	size = OBEX_ObjectReadStream(obex, obj, &buffer);
 	if (size < 0) {
 		error("Error on OBEX stream");
@@ -437,6 +616,7 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 	guint hlen;
 	guint8 hi;
 	guint64 free;
+	int ret;
 
 	os = OBEX_GetUserData(obex);
 
@@ -489,13 +669,17 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 			break;
 
 		case OBEX_HDR_BODY:
-			os->size = -1;
+			if (os->size < 0)
+				os->size = OBJECT_SIZE_UNKNOWN;
 			break;
 
 		case OBEX_HDR_LENGTH:
 			os->size = hd.bq4;
 			debug("OBEX_HDR_LENGTH: %d", os->size);
 			break;
+		case OBEX_HDR_TIME:
+			os->time = parse_iso8610((const gchar *) hd.bs, hlen);
+			break;
 		}
 	}
 
@@ -509,12 +693,30 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 		return FALSE;
 	}
 
-	if (!os->cmds->chkput)
+	if (!os->cmds || !os->cmds->chkput)
 		goto done;
 
-	if (os->cmds->chkput(obex, obj) < 0) {
+	ret = os->cmds->chkput(obex, obj);
+	switch (ret) {
+	case 0:
+		break;
+	case -EINVAL:
+		OBEX_ObjectSetRsp(obj, OBEX_RSP_BAD_REQUEST,
+				OBEX_RSP_BAD_REQUEST);
+		return FALSE;
+	case -EPERM:
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 		return FALSE;
+	default:
+		debug("Unhandled chkput error: %d", ret);
+		OBEX_ObjectSetRsp(obj, OBEX_RSP_INTERNAL_SERVER_ERROR,
+				OBEX_RSP_INTERNAL_SERVER_ERROR);
+		return FALSE;
+	}
+
+	if (os->size == OBJECT_SIZE_DELETE || os->size == OBJECT_SIZE_UNKNOWN) {
+		debug("Got a PUT without a Length");
+		goto done;
 	}
 
 	if (fstatvfs(os->fd, &buf) < 0) {
@@ -540,7 +742,10 @@ done:
 
 static void cmd_put(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 {
-	if (!os->cmds->put) {
+	if (!os->cmds) {
+		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
+		return;
+	} else if (!os->cmds->put) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
 				OBEX_RSP_NOT_IMPLEMENTED);
 		return;
@@ -567,9 +772,15 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 
 	switch (evt) {
 	case OBEX_EV_PROGRESS:
-		emit_transfer_progress(os->cid, os->size, os->offset);
+		/* Just emit progress for Object Push */
+		if (os->target == NULL)
+			emit_transfer_progress(os->cid, os->size, os->offset);
 		break;
 	case OBEX_EV_ABORT:
+		os->aborted = TRUE;
+		if (os->target == NULL)
+			emit_transfer_completed(os->cid, FALSE);
+		os_reset_session(os);
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_SUCCESS, OBEX_RSP_SUCCESS);
 		break;
 	case OBEX_EV_REQDONE:
@@ -579,21 +790,17 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 			break;
 		case OBEX_CMD_PUT:
 		case OBEX_CMD_GET:
-			emit_transfer_completed(os->cid,
-						os->offset == os->size);
-			if (os->fd >= 0) {
-				close(os->fd);
-				os->fd = -1;
-			}
-			g_free(os->buf);
-			os->buf = NULL;
-			os->offset = 0;
+			os_session_mark_aborted(os);
+			if (os->target == NULL)
+				emit_transfer_completed(os->cid, !os->aborted);
+			os_reset_session(os);
 			break;
 		default:
 			break;
 		}
 		break;
 	case OBEX_EV_REQHINT:
+		os->cmd = cmd;
 		switch (cmd) {
 		case OBEX_CMD_PUT:
 			os->checked = FALSE;
@@ -614,7 +821,7 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 	case OBEX_EV_REQCHECK:
 		switch (cmd) {
 		case OBEX_CMD_PUT:
-			if (os->cmds->put)
+			if (os->cmds && os->cmds->put)
 				check_put(obex, obj);
 			break;
 		default:
@@ -645,14 +852,23 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 		}
 		break;
 	case OBEX_EV_STREAMAVAIL:
-		if (obex_read(os, obex, obj) < 0) {
-			debug("error obex_read()");
-			OBEX_CancelRequest(obex, 1);
+		switch (obex_read_stream(os, obex, obj)) {
+		case 0:
+			break;
+		case -EPERM:
+			OBEX_ObjectSetRsp(obj,
+				OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
+			break;
+		default:
+			OBEX_ObjectSetRsp(obj,
+				OBEX_RSP_INTERNAL_SERVER_ERROR,
+				OBEX_RSP_INTERNAL_SERVER_ERROR);
+			break;
 		}
 
 		break;
 	case OBEX_EV_STREAMEMPTY:
-		obex_write(os, obex, obj);
+		obex_write_stream(os, obex, obj);
 		break;
 	case OBEX_EV_LINKERR:
 		break;
@@ -667,6 +883,15 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 	}
 }
 
+void server_free(struct server *server)
+{
+	g_free(server->name);
+	g_free(server->folder);
+	g_free(server->capability);
+	g_free(server->devnode);
+	g_free(server);
+}
+
 static void obex_handle_destroy(gpointer user_data)
 {
 	struct obex_session *os;
@@ -674,17 +899,33 @@ static void obex_handle_destroy(gpointer user_data)
 
 	os = OBEX_GetUserData(obex);
 
-	/* Got an error during a transfer. */
-	if (os->fd >= 0)
-		emit_transfer_completed(os->cid, os->offset == os->size);
+	if (os->target == NULL) {
+		/* Got an error during a transfer. */
+		if (os->fd >= 0)
+			emit_transfer_completed(os->cid, os->offset == os->size);
 
-	unregister_transfer(os->cid);
+		unregister_transfer(os->cid);
+	} else {
+		emit_session_removed(os->cid);
+		unregister_session(os->cid);
+	}
 
 	obex_session_free(os);
 
 	OBEX_Cleanup(obex);
 }
 
+static gboolean tty_reinit(gpointer data)
+{
+	struct server *server = data;
+
+	tty_init(server->services, server->folder, server->capability, server->devnode);
+
+	server_free(server);
+
+	return FALSE;
+}
+
 static gboolean obex_handle_input(GIOChannel *io,
 				GIOCondition cond, gpointer user_data)
 {
@@ -693,8 +934,17 @@ static gboolean obex_handle_input(GIOChannel *io,
 	if (cond & G_IO_NVAL)
 		return FALSE;
 
-	if (cond & (G_IO_HUP | G_IO_ERR))
+	if (cond & (G_IO_HUP | G_IO_ERR)) {
+		struct obex_session *os;
+
+		error("HUP");
+
+		os = OBEX_GetUserData(obex);
+		if (os->server->devnode)
+			g_idle_add(tty_reinit, os->server);
+
 		return FALSE;
+	}
 
 	if (OBEX_HandleInput(obex, 1) < 0) {
 		error("Handle input error");
@@ -712,26 +962,18 @@ gint obex_session_start(gint fd, struct server *server)
 	gint ret;
 
 	os = g_new0(struct obex_session, 1);
-	switch (server->service) {
-	case OBEX_OPUSH:
-		os->target = NULL;
+
+	os->target = NULL;
+
+	if (server->services & OBEX_OPP)
 		os->cmds = &opp;
-		break;
-	case OBEX_FTP:
-		os->target = FTP_TARGET;
-		os->cmds = &ftp;
-		break;
-	default:
-		g_free(os);
-		debug("Invalid OBEX server");
-		return -EINVAL;
-	}
 
 	os->current_folder = g_strdup(server->folder);
 	os->server = server;
-	os->rx_mtu = RX_MTU;
-	os->tx_mtu = TX_MTU;
+	os->rx_mtu = server->rx_mtu ? server->rx_mtu : DEFAULT_RX_MTU;
+	os->tx_mtu = server->tx_mtu ? server->tx_mtu : DEFAULT_TX_MTU;
 	os->fd = -1;
+	os->size = OBJECT_SIZE_DELETE;
 
 	obex = OBEX_Init(OBEX_TRANS_FD, obex_event, 0);
 	if (!obex) {
@@ -754,6 +996,7 @@ gint obex_session_start(gint fd, struct server *server)
 	g_io_add_watch_full(io, G_PRIORITY_DEFAULT,
 			G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
 			obex_handle_input, obex, obex_handle_destroy);
+	g_io_channel_set_close_on_unref(io, TRUE);
 	g_io_channel_unref(io);
 
 	return 0;
